# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Classes to help developers avoid code duplication when writing tests for pyiron.

TODO:
    This is just a direct copy of the same file in pyiron_base. As soon as sub-module Project objects
    actually *update* the Project class (e.g. with new functionality on the create attribute) then this
    module can be removed and the module directly from base can be used.
"""

import unittest
from abc import ABC
from inspect import getfile
from os import remove
from os.path import join, split

from pyiron_atomistics.project import Project

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Mar 23, 2021"


class TestWithProject(unittest.TestCase, ABC):
    """
    Tests that start and remove a project for their suite.
    """

    @classmethod
    def setUpClass(cls):
        cls.project_path = getfile(cls)[:-3].replace("\\", "/")
        cls.file_location, cls.project_name = split(cls.project_path)
        cls.project = Project(cls.project_path)

    @classmethod
    def tearDownClass(cls):
        cls.project.remove(enable=True)
        try:
            remove(join(cls.file_location, "pyiron.log"))
        except FileNotFoundError:
            pass


class TestWithCleanProject(TestWithProject, ABC):
    """
    Tests that start and remove a project for their suite, and remove jobs from the project for each test.
    """

    def tearDown(self):
        self.project.remove_jobs(recursive=True, silently=True)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Alternative structure container that stores them in flattened arrays.
"""

from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pyiron_base import FlattenedStorage
from pyiron_snippets.import_alarm import ImportAlarm
from structuretoolkit.common.error import SymmetryError

import pyiron_atomistics.atomistics.structure.has_structure as pa_has_structure
from pyiron_atomistics.atomistics.structure.atom import Atom
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.neighbors import NeighborsTrajectory

with ImportAlarm(
    "Some plotting functionality requires the seaborn library."
) as seaborn_alarm:
    import seaborn as sns


class StructureStorage(FlattenedStorage, pa_has_structure.HasStructure):
    """
    Class that can write and read lots of structures from and to hdf quickly.

    This is done by storing positions, cells, etc. into large arrays instead of writing every structure into a new
    group.  Structures are stored together with an identifier that should be unique.  The class can be initialized with
    the number of structures and the total number of atoms in all structures, but re-allocates memory as necessary when
    more (or larger) structures are added than initially anticipated.

    You can add structures and a human-readable name with :meth:`.add_structure()`.

    >>> container = StructureStorage()
    >>> container.add_structure(Atoms(...), "fcc")
    >>> container.add_structure(Atoms(...), "hcp")
    >>> container.add_structure(Atoms(...), "bcc")

    Accessing stored structures works with :meth:`.get_strucure()`.  You can either pass the identifier you passed
    when adding the structure or the numeric index

    >>> container.get_structure(frame=0) == container.get_structure(frame="fcc")
    True

    Custom arrays may also be defined on the container

    >>> container.add_array("energy", shape=(), dtype=np.float64, fill=-1, per="chunk")

    (chunk means structure in this case, see below and :class:`.FlattenedStorage`)

    You can then pass arrays of the corresponding shape to :meth:`add_structure()`

    >>> container.add_structure(Atoms(...), "grain_boundary", energy=3.14)

    Saved arrays are accessed with :meth:`.get_array()`

    >>> container.get_array("energy", 3)
    3.14
    >>> container.get_array("energy", 0)
    -1

    It is also possible to use the same names in :meth:`.get_array()` as in :meth:`.get_structure()`.

    >>> container.get_array("energy", 0) == container.get_array("energy", "fcc")
    True

    The length of the container is the number of structures inside it.

    >>> len(container)
    4

    Each structure corresponds to a chunk in :class:`.FlattenedStorage` and each atom to an element.  By default the
    following arrays are defined for each structure:
        - identifier    shape=(),    dtype=str,          per chunk; human readable name of the structure
        - cell          shape=(3,3), dtype=np.float64,   per chunk; cell shape
        - pbc           shape=(3,),  dtype=bool          per chunk; periodic boundary conditions
        - symbols:      shape=(),    dtype=str,          per element; chemical symbol
        - positions:    shape=(3,),  dtype=np.float64,   per element: atomic positions
    If a structure has spins/magnetic moments defined on its atoms these will be saved in a per atom array as well.  In
    that case, however all structures in the container must either have all collinear spins or all non-collinear spins.
    """

    def __init__(self, num_atoms=1, num_structures=1):
        """
        Create new structure container.

        Args:
            num_atoms (int): total number of atoms across all structures to pre-allocate
            num_structures (int): number of structures to pre-allocate
        """
        super().__init__(num_elements=num_atoms, num_chunks=num_structures)
        self._element_cache = None
        self._plots = None

    def _init_arrays(self):
        super()._init_arrays()
        # 2 character unicode array for chemical symbols
        self._per_element_arrays["symbols"] = np.full(
            self._num_elements_alloc, "XX", dtype=np.dtype("U2")
        )
        self._per_element_arrays["positions"] = np.empty((self._num_elements_alloc, 3))

        self._per_chunk_arrays["cell"] = np.empty((self._num_chunks_alloc, 3, 3))
        self._per_chunk_arrays["pbc"] = np.empty(
            (self._num_elements_alloc, 3), dtype=bool
        )

    @property
    def symbols(self):
        """:meta private:"""
        return self._per_element_arrays["symbols"]

    @property
    def positions(self):
        """:meta private:"""
        return self._per_element_arrays["positions"]

    @property
    def start_index(self):
        """:meta private:"""
        return self._per_chunk_arrays["start_index"]

    @property
    def length(self):
        """:meta private:"""
        return self._per_chunk_arrays["length"]

    @property
    def identifier(self):
        """:meta private:"""
        return self._per_chunk_arrays["identifier"]

    @property
    def cell(self):
        """:meta private:"""
        return self._per_chunk_arrays["cell"]

    @property
    def pbc(self):
        """:meta private:"""
        return self._per_chunk_arrays["pbc"]

    def set_array(self, name, frame, value):
        """
        Add array for given structure.

        Works for per chunk and per element arrays.

        Args:
            name (str): name of array to set
            frame (int, str): selects structure to set, as in :meth:`.get_strucure()`
            value: value (for per chunk) or array of values (for per element); type and shape as per :meth:`.hasarray()`.

        Raises:
            `KeyError`: if array with name does not exists
        """
        super().set_array(name, frame, value)
        # invalidate cache when writing to symbols; could be smarter by checking if value is not in in cache but w/e
        if name == "symbols":
            self._element_cache = None

    def get_elements(self) -> List[str]:
        """
        Return a list of chemical elements present in the storage.

        Returns:
            :class:`list`: list of unique elements as strings of chemical symbols
        """
        if self._element_cache is None:
            self._element_cache = list(np.unique(self._per_element_arrays["symbols"]))
        return self._element_cache

    def add_structure(self, structure, identifier=None, **arrays):
        """
        Add a new structure to the container.

        Additional keyword arguments given specify additional arrays to store for the structure.  If an array with the
        given keyword name does not exist yet, it will be added to the container.

        >>> container = StructureStorage()
        >>> container.add_structure(Atoms(...), identifier="A", energy=3.14)
        >>> container.get_array("energy", 0)
        3.14

        If the first axis of the extra array matches the length of the given structure, it will be added as an per atom
        array, otherwise as an per structure array.

        >>> structure = Atoms(...)
        >>> container.add_structure(structure, identifier="B", forces=len(structure) * [[0,0,0]])
        >>> len(container.get_array("forces", 1)) == len(structure)
        True

        Reshaping the array to have the first axis be length 1 forces the array to be set as per structure array.  That
        axis will then be stripped.

        >>> container.add_structure(Atoms(...), identifier="C", pressure=np.eye(3)[np.newaxis, :, :])
        >>> container.get_array("pressure", 2).shape
        (3, 3)

        Args:
            structure (:class:`.Atoms`): structure to add
            identifier (str, optional): human-readable name for the structure, if None use current structre index as
                                        string
            **kwargs: additional arrays to store for structure
        """

        if structure.has("initial_magmoms"):
            arrays["spins"] = structure.spins
        if "selective_dynamics" in structure.get_tags():
            arrays["selective_dynamics"] = structure.selective_dynamics

        self.add_chunk(
            len(structure),
            identifier=identifier,
            symbols=np.array(structure.symbols),
            positions=structure.positions,
            cell=[structure.cell.array],
            pbc=[structure.pbc],
            **arrays,
        )

    def _translate_frame(self, frame):
        try:
            return self.find_chunk(frame)
        except KeyError:
            raise KeyError(f"No structure named {frame}.") from None

    def _get_structure(self, frame=-1, wrap_atoms=True):
        symbols = self.get_array("symbols", frame)
        elements = [e for e in self.get_elements() if e in symbols]
        index_map = {e: i for i, e in enumerate(elements)}
        try:
            magmoms = self.get_array("spins", frame)
        except KeyError:
            # not all structures have spins saved on them
            magmoms = None
        structure = Atoms(
            species=[Atom(e).element for e in elements],
            indices=[index_map[e] for e in symbols],
            positions=self.get_array("positions", frame),
            cell=self.get_array("cell", frame),
            pbc=self.get_array("pbc", frame),
            magmoms=magmoms,
        )
        if self.has_array("selective_dynamics"):
            structure.add_tag(selective_dynamics=[True, True, True])
            selective_dynamics = self.get_array("selective_dynamics", frame)
            for i, d in enumerate(selective_dynamics):
                structure.selective_dynamics[i] = d.tolist()
        return structure

    def _number_of_structures(self):
        return len(self)

    def _get_hdf_group_name(self):
        return "structures"

    @property
    def plot(self):
        """
        Accessor for :class:`.StructurePlots` instance using these structures.
        """
        if self._plots is None:
            self._plots = StructurePlots(self)
        return self._plots


class StructurePlots:
    """
    Simple interface to plot various properties of structures.
    """

    @seaborn_alarm
    def __init__(self, store: StructureStorage):
        self._store = store
        self._neigh = None

    def atoms(self):
        """
        Plot a histogram of the number of atoms in each structure.
        """
        length = self._store["length"]
        lo = length.min()
        hi = length.max()
        # make the bins fall in between whole numbers and include hi
        plt.hist(length, bins=np.arange(lo, hi + 2) - 0.5)
        plt.xlabel("#Atoms")
        plt.ylabel("Count")

    def cell(self, angle_in_degrees=True):
        """
        Plot histograms of cell parameters.

        Plotted are atomic volume, density, cell vector lengths and cell vector angles in separate subplots all on a
        log-scale.

        Args:
            angle_in_degrees (bool): whether unit for angles is degree or radians

        Returns:
            `DataFrame`: contains the plotted information in the columns:
                            - a: length of first vector
                            - b: length of second vector
                            - c: length of third vector
                            - alpha: angle between first and second vector
                            - beta: angle between second and third vector
                            - gamma: angle between third and first vector
                            - V: volume of the cell
                            - N: number of atoms in the cell
        """
        N = self._store.get_array("length")
        C = self._store.get_array("cell")

        def get_angle(cell, idx=0):
            return np.arccos(
                np.dot(cell[idx], cell[(idx + 1) % 3])
                / np.linalg.norm(cell[idx])
                / np.linalg.norm(cell[(idx + 1) % 3])
            )

        def extract(n, c):
            return {
                "a": np.linalg.norm(c[0]),
                "b": np.linalg.norm(c[1]),
                "c": np.linalg.norm(c[2]),
                "alpha": get_angle(c, 0),
                "beta": get_angle(c, 1),
                "gamma": get_angle(c, 2),
            }

        df = pd.DataFrame([extract(n, c) for n, c in zip(N, C)])
        df["V"] = np.linalg.det(C)
        df["N"] = N
        if angle_in_degrees:
            df["alpha"] = np.rad2deg(df["alpha"])
            df["beta"] = np.rad2deg(df["beta"])
            df["gamma"] = np.rad2deg(df["gamma"])

        plt.subplot(1, 4, 1)
        plt.title("Atomic Volume")
        plt.hist(df.V / df.N, bins=20, log=True)
        plt.xlabel(r"$V$ [$\AA^3$]")

        plt.subplot(1, 4, 2)
        plt.title("Density")
        plt.hist(df.N / df.V, bins=20, log=True)
        plt.xlabel(r"$\rho$ [$\AA^{-3}$]")

        plt.subplot(1, 4, 3)
        plt.title("Lattice Vector Lengths")
        plt.hist([df.a, df.b, df.c], log=True)
        plt.xlabel(r"$a,b,c$ [$\AA$]")

        plt.subplot(1, 4, 4)
        plt.title("Lattice Vector Angles")
        plt.hist([df.alpha, df.beta, df.gamma], log=True)
        if angle_in_degrees:
            label = r"$\alpha,\beta,\gamma$ [°]"
        else:
            label = r"$\alpha,\beta,\gamma$ [rad]"
        plt.xlabel(label)

        return df

    def _calc_spacegroups(self, symprec=1e-3):
        """
        Calculate space groups of all structures.

        Args:
            symprec (float): symmetry precision given to spglib

        Returns:
            DataFrame: contains columns 'crystal_system' (str) and 'space_group' (int) for each structure
        """

        def get_crystal_system(num):
            if num in range(1, 3):
                return "triclinic"
            elif num in range(3, 16):
                return "monoclinic"
            elif num in range(16, 75):
                return "orthorhombic"
            elif num in range(75, 143):
                return "tetragonal"
            elif num in range(143, 168):
                return "trigonal"
            elif num in range(168, 195):
                return "hexagonal"
            elif num in range(195, 231):
                return "cubic"

        def extract(s):
            try:
                spg = s.get_symmetry(symprec=symprec).spacegroup["Number"]
            except SymmetryError:
                spg = 1
            return {"space_group": spg, "crystal_system": get_crystal_system(spg)}

        return pd.DataFrame(map(extract, self._store.iter_structures()))

    def spacegroups(self, symprec=1e-3):
        """
        Plot histograms of space groups and crystal systems.

        Spacegroups and crystal systems are plotted in separate subplots.

        Args:
            symprec (float): precision of the symmetry search (passed to spglib)

        Returns:
            DataFrame: contains two columns "space_group", "crystal_system"
                       for each structure
        """

        df = self._calc_spacegroups(symprec=symprec)
        plt.subplot(1, 2, 1)
        plt.hist(df.space_group, bins=230)
        plt.xlabel("Space Group")

        plt.subplot(1, 2, 2)
        l, h = np.unique(df.crystal_system, return_counts=True)
        sort_key = {
            "triclinic": 1,
            "monoclinic": 3,
            "orthorhombic": 16,
            "tetragonal": 75,
            "trigonal": 143,
            "hexagonal": 168,
            "cubic": 195,
        }
        I = np.argsort([sort_key[ll] for ll in l])
        plt.bar(l[I], h[I])
        plt.xlabel("Crystal System")
        plt.xticks(rotation=35)
        return df

    def _calc_neighbors(self, num_neighbors):
        """
        Calculate the neighbor information with additional caching.

        If 'distances' and 'shells' are provided in the underlying store, they are returned directly without checking
        `num_neighbors`.

        If they are not provided there, they are calculated here and cached.
        Recalculation happens when a different `num_neighbors` is provided than in a previous call or the underlying
        store changes.

        If `num_neighbors` is `None` on the first call, the default is 36.

        Returns:
            dict: with keys 'distances' and 'shells' containing the respective flattened arrays from
            :meth:`.Atoms.get_neighbors`.
        """
        if self._store.has_array("distances") and self._store.has_array("shells"):
            return {
                "distances": self._store["distances"],
                "shells": self._store["shells"],
            }
        # check that _store and _neigh are still consistent
        if (
            self._neigh is None
            or len(self._store) != len(self._neigh)
            or (
                num_neighbors is None
                or self._neigh.has_array("distances")["shape"][0] != num_neighbors
            )
        ):
            if num_neighbors is None:
                num_neighbors = 36
            self._neigh = FlattenedStorage()
            neigh_traj = NeighborsTrajectory(
                has_structure=self._store,
                num_neighbors=num_neighbors,
                store=self._neigh,
            )
        return {
            "distances": self._neigh["distances"],
            "shells": self._neigh["shells"],
        }

    def coordination(self, num_shells=4, log=True, num_neighbors=None):
        """
        Plot histogram of coordination in neighbor shells.

        Computes one histogram of the number of neighbors in each neighbor shell up to `num_shells` and then plots them
        together.

        If the underlying :class:`.StructureStorage` has a 'shells' array defined it is used, if not it is calculated on
        the fly.

        Args:
            num_shells (int): maximum shell to plot
            num_neighbors (int): maximum number of neighbors to calculate, when 'shells' is not defined in storage,
                                 default is the value from the previous call or 36
            log (float): plot histogram values on a log scale
        """
        neigh = self._calc_neighbors(num_neighbors=num_neighbors)
        shells = neigh["shells"]

        shell_index = (
            shells[np.newaxis, :, :]
            == np.arange(1, num_shells + 1)[:, np.newaxis, np.newaxis]
        )
        neigh_count = shell_index.sum(axis=-1)
        ticks = np.arange(neigh_count.min(), neigh_count.max() + 1)
        plt.hist(
            neigh_count.T,
            bins=ticks - 0.5,
            log=True,
            label=[f"{i}." for i in range(1, num_shells + 1)],
        )
        plt.xticks(ticks)
        plt.xlabel("Number of Neighbors")
        plt.legend(title="Shell")
        plt.title("Neighbor Coordination in Shells")

    def distances(
        self,
        bins: int = 50,
        num_neighbors: int = None,
        normalize: bool = False,
    ):
        """
        Plot a histogram of the neighbor distances.

        Setting `normalize` plots the radial distribution function.

        Args:
            bins (int): number of bins
            num_neighbors (int): maximum number of neighbors to calculate, when 'shells' or 'distances' are not defined in storage
                                 default is the value from the previous call or 36
            normalize (bool): normalize the distribution by the surface area of
                              the radial bin, 4pi r^2
        """
        neigh = self._calc_neighbors(num_neighbors=num_neighbors)
        distances = neigh["distances"].flatten()

        if normalize:
            plt.hist(
                distances,
                bins=bins,
                weights=1 / (4 * np.pi * distances**2),
            )
            plt.ylabel("Neighbor density [$\mathrm{\AA}^{-2}$]")
        else:
            plt.hist(distances, bins=bins)
            plt.ylabel("Neighbor count")
        plt.xlabel(r"Distance [$\mathrm{\AA}$]")

    def shell_distances(self, num_shells=4, num_neighbors=None):
        """
        Plot a violin plot of the neighbor distances in shells up to `num_shells`.

        Args:
            num_shells (int): maximum shell to plot
            num_neighbors (int): maximum number of neighbors to calculate, when 'shells' or 'distances' are not defined in storage
                                 default is the value from the previous call or 36
        """
        neigh = self._calc_neighbors(num_neighbors=num_neighbors)
        shells = neigh["shells"]
        distances = neigh["distances"]

        R = distances.flatten()
        S = shells.ravel()
        d = pd.DataFrame(
            {"distance": R[S < num_shells + 1], "shells": S[S < num_shells + 1]}
        )
        sns.violinplot(y=d.shells, x=d.distance, scale="width", orient="h")
        plt.xlabel(r"Distance [$\AA$]")
        plt.ylabel("Shell")

    def concentration(self, elements: List[str] = None, **kwargs) -> pd.DataFrame:
        """
        Plot histograms of the concentrations in each structure.

        Args:
            elements (list of str): elements to plot the histograms for; default is for all elements in the container
            **kwargs: passed through to `seaborn.histplot`

        Returns:
            `pandas.DataFrame`: table of concentrations in each structure; column headers are the element names
        """
        if elements is not None:
            for elem in elements:
                if elem not in self._store.get_elements():
                    raise ValueError(f"Element {elem} not present in storage!")
        else:
            elements = self._store.get_elements()

        df = pd.DataFrame(
            [
                {elem: sum(elem == sym) / len(sym) for elem in elements}
                for sym in self._store.get_array_ragged("symbols")
            ]
        )

        sns.histplot(
            data=df.melt(var_name="element", value_name="concentration"),
            x="concentration",
            hue="element",
            multiple="dodge",
            **kwargs,
        )

        return df



from typing import List, Union

from pyiron_atomistics.atomistics.structure.atoms import Atoms, pymatgen_to_pyiron
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage


class MPQueryResults(HasStructure):
    def __init__(self, results):
        """

        Args:
            docs (list of dicts): query results from Materials Project, should be obtained with use_document_model=False
        """
        self._results = results
        self._material_ids = [r["material_id"] for r in results]

    def _translate_frame(self, frame):
        try:
            return self._material_ids.index(frame)
        except ValueError:
            raise KeyError(f"material id {frame} not among results!") from None

    def __iter__(self):
        yield from self.iter_structures()

    def _get_structure(self, frame, wrap_atoms=True):
        return pymatgen_to_pyiron(self._results[frame]["structure"])

    def _number_of_structures(self):
        return len(self._results)

    def to_list(self):
        """
        Get a list of queried structures.

        Returns:
            list: structures
        """
        return [pymatgen_to_pyiron(r["structure"]) for r in self._results]

    def to_storage(self):
        """
        Get a StructureStorage of queried structures.

        The materials project id is used as the identifier.

        Returns:
            :class:`~.StructureStorage`: structures
        """
        store = StructureStorage()
        for i, structure in enumerate(self):
            store.add_structure(structure, identifier=self._material_ids[i])
        return store


class MaterialsProjectFactory:
    """
    Convenience interface to the Materials Project Structure Database.

    Usage is only possible with an API key obtained from the Materials Project.  To do this, create an account with
    them, login and access `this webpage <https://next-gen.materialsproject.org/api#api-key>`.

    Once you have a key, either pass it as the `api_key` parameter in the methods of this object or export an
    environment variable, called `MP_API_KEY`, in your shell setup.
    """

    @staticmethod
    def search(
        chemsys: Union[str, List[str]], api_key=None, **kwargs
    ) -> MPQueryResults:
        """
        Search the database for all structures matching the given query.

        Note that `chemsys` takes distint values for unaries, binaries and so!  A query with `chemsys=["Fe", "O"]` will
        return iron structures and oxygen structures, but no iron oxide structures.  Similarily `chemsys=["Fe-O"]` will
        not return unary structures.

        All keyword arguments for filtering from the original API are supported.  See the
        `original docs <https://docs.materialsproject.org/downloading-data/using-the-api>`_ for them.

        Search for all iron structures:

        >>> pr = Project(...)
        >>> irons = pr.create.structure.materialsproject.search("Fe")
        >>> irons.number_of_structures
        10

        The returned :class:`~.MPQueryResults` object implements :class:`~.HasStructure` and can be accessed with the
        material ids as a short-hand

        >>> irons.get_structure(1) == irons.get_structure('mp-13')
        True

        Search for all structures with Al, Li that are on the T=0 convex hull:

        >>> alli = pr.create.structure.materialsproject.search(['Al', 'Li', 'Al-Li'], is_stable=True)
        >>> len(alli)
        6

        Args:
            chemsys (str, list of str): confine search to given elements; either an element symbol or multiple element
            symbols seperated by dashes; if a list of strings is given return structures matching either of them
            api_key (str, optional): if your API key is not exported in the environment flag MP_API_KEY, pass it here
            **kwargs: passed verbatim to :meth:`mp_api.MPRester.summary.search` to further filter the results

        Returns:
            :class:`~.MPQueryResults`: resulting structures from the query
        """
        from mp_api.client import MPRester

        rest_kwargs = {
            "use_document_model": False,  # returns results as dictionaries
            "include_user_agent": True,  # send some additional software version info to MP
        }
        if api_key is not None:
            rest_kwargs["api_key"] = api_key
        with MPRester(**rest_kwargs) as mpr:
            results = mpr.summary.search(
                chemsys=chemsys, **kwargs, fields=["structure", "material_id"]
            )
        return MPQueryResults(results)

    @staticmethod
    def by_id(
        material_id: Union[str, int],
        final: bool = True,
        conventional_unit_cell: bool = False,
        api_key=None,
    ) -> Union[Atoms, List[Atoms]]:
        """
        Retrieve a structure by material id.

        This is how you would ask for the iron ground state:

        >>> pr = Project(...)
        >>> pr.create.structure.materialsproject.by_id('mp-13')
        Fe: [0. 0. 0.]
        tags:
            spin: [(0: 2.214)]
        pbc: [ True  True  True]
        cell:
        Cell([[2.318956, 0.000185, -0.819712], [-1.159251, 2.008215, -0.819524], [2.5e-05, 0.000273, 2.459206]])


        Args:
            material_id (str): the id assigned to a structure by the materials project
            api_key (str, optional): if your API key is not exported in the environment flag MP_API_KEY, pass it here
            final (bool, optional): if set to False, returns the list of initial structures,
            else returns the final structure. (Default is True)
            conventional_unit_cell (bool, optional): if set to True, returns the standard conventional unit cell.
            (Default is False)

        Returns:
            :class:`~.Atoms`: requested final structure if final is True
            list of :class:~.Atoms`:  a list of initial (pre-relaxation) structures if final is False

        Raises:
            ValueError: material id does not exist
        """
        from mp_api.client import MPRester

        rest_kwargs = {
            "include_user_agent": True,  # send some additional software version info to MP
        }
        if api_key is not None:
            rest_kwargs["api_key"] = api_key
        with MPRester(**rest_kwargs) as mpr:
            if final:
                return pymatgen_to_pyiron(
                    mpr.get_structure_by_material_id(
                        material_id=material_id,
                        final=final,
                        conventional_unit_cell=conventional_unit_cell,
                    )
                )
            else:
                return [
                    pymatgen_to_pyiron(mpr_structure)
                    for mpr_structure in (
                        mpr.get_structure_by_material_id(
                            material_id=material_id,
                            final=final,
                            conventional_unit_cell=conventional_unit_cell,
                        )
                    )
                ]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from pyiron_base import DataContainer

from pyiron_atomistics.atomistics.job.interactivewrapper import (
    InteractiveWrapper,
    ReferenceJobOutput,
)

__author__ = "Osamu Waseda"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH "
    "- Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Osamu Waseda"
__email__ = "waseda@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class ARTInteractive(object):
    def __init__(self, art_id, direction, gamma=0.1, fix_layer=False, non_art_id=None):
        if int(art_id) != art_id or art_id < 0:
            raise ValueError("art_id must be a posive integer")
        if len(direction) != 3 or np.isclose(np.linalg.norm(direction), 0):
            raise ValueError("direction must be a finite 3d vector")
        if gamma < 0:
            raise ValueError("gamma must be a positive float")
        if fix_layer and non_art_id is not None:
            raise ValueError("fix_layer and non_art_id cannot be set at the same time")
        self.art_id = art_id
        self.direction = direction
        self.gamma = gamma
        self.non_art_id = non_art_id
        if non_art_id is not None:
            self.non_art_id = np.array([non_art_id]).flatten()
        self.fix_layer = fix_layer

    @property
    def _R(self):
        value = np.array(self.direction)
        value = value / np.linalg.norm(value)
        return np.outer(value, value)

    def get_forces(self, f_in):
        f = np.array(f_in)
        if len(f.shape) == 2:
            f = np.array([f])
        if self.non_art_id is None:
            self.non_art_id = np.arange(len(f[0])) != self.art_id
        f_art = (1.0 + self.gamma) * np.einsum("ij,nj->ni", self._R, f[:, self.art_id])
        if self.fix_layer:
            f[:, self.non_art_id] = np.einsum(
                "nmj,ij->nmi", f[:, self.non_art_id], np.identity(3) - self._R
            )
        else:
            f[:, self.non_art_id] += f_art[:, np.newaxis, :] / np.sum(
                self.non_art_id != False
            )
        f[:, self.art_id] -= f_art
        return f.reshape(np.array(f_in).shape)


class ART(InteractiveWrapper):
    """
    Apply an artificial force according to the Activation Relaxation Technique (ART)
    DOI:https://doi.org/10.1103/PhysRevE.57.2419

    The applied force f_art is calculated from the original force f by:

    f_art = f-(1+gamma)*np.dot(n,f)*n

    where gamma is a parameter to be determined in the input (default: 0.1) and n is
    the direction along which the force is reversed (3d-vector). In order to homogenize
    the total force in the system, f_art is distributed among atoms specified by
    non_art_id (default: all atoms), or if fix_layer is defined, the forces acting on
    all the other atoms along the direction n are cancelled. Note: Since the energy
    is not compatible with the forces anymore, structure optimization methods which
    rely on the energy variation (such as conjugate gradient) cannot/should not be used.

    Input:
        - art_id (int): atom id on which ART force is applied
        - direction (list/numpy.ndarray): direction along which force is reversed
        - gamma (float): prefactor for force inversion. v.s.
        - non_art_id (list/numpy.ndarray): list of atoms to be used for the force cancellation
        - fix_layer (bool): whether or not to fix all other atoms on the layer perpendicular
            to the direction along which force is reversed.

    Example:

        # Structure creation
        >>> vacancy_position = structure.positions[0]
        >>> del structure[0]
        >>> neighbors = structure.get_neighborhood(vacancy_position)
        >>> direction = neighbors.vecs[0]
        >>> art_id = neighbors.indices[0]
        >>> structure.positions[art_id] -= 0.5*direction

        # Job creation
        >>> some_atomistic_job.structure = structure
        >>> art = ART(job_name='art')
        >>> art.ref_job = some_atomistic_job
        >>> art.input.art_id = art_id
        >>> art.input.direction = direction
        >>> some_minimizer.ref_job = art
        >>> some_minimizer.run()


    """

    def __init__(self, project, job_name):
        super(ART, self).__init__(project, job_name)

        self.input = DataContainer(table_name="custom_dict")
        self.input.gamma = 0.1
        self.input.fix_layer = False
        self.input.non_art_id = None
        self.input.art_id = None
        self.input.direction = None
        self.output = ARTIntOutput(job=self)
        self.server.run_mode.interactive = True
        self._interactive_interface = None
        self._art = None

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        self.input.read_only = True

    @property
    def art(self):
        if self._art is None:
            self._art = ARTInteractive(
                art_id=self.input.art_id,
                direction=self.input.direction,
                gamma=self.input.gamma,
                fix_layer=self.input.fix_layer,
                non_art_id=self.input.non_art_id,
            )
        return self._art

    def run_if_interactive(self):
        self._logger.debug("art status: " + str(self.status))
        if not self.status.running:
            self.ref_job_initialize()
        self.status.running = True
        if self.ref_job.server.run_mode.interactive:
            self.ref_job.run()
        else:
            self.ref_job.run(run_again=True)
        self._logger.debug("art status: " + str(self.status))

    def interactive_forces_getter(self):
        return self.art.get_forces(self.ref_job.output.forces[-1])

    def validate_ready_to_run(self):
        """
        check whether art_id and direction are set
        """
        if self.input.art_id is None or self.input.direction is None:
            raise AssertionError("art_id and/or direction not set")

    def interactive_close(self):
        self.status.collect = True
        if self.ref_job.server.run_mode.interactive:
            self.ref_job.interactive_close()
        self.status.finished = True


class ARTIntOutput(ReferenceJobOutput):
    def __init__(self, job):
        super(ARTIntOutput, self).__init__(job=job)

    @property
    def forces(self):
        return self._job.art.get_forces(self._job.ref_job.output.forces)



import posixpath
from shutil import copyfile

import numpy as np
import scipy.constants
from pyiron_base import DataContainer

BOHR_TO_ANGSTROM = (
    scipy.constants.physical_constants["Bohr radius"][0] / scipy.constants.angstrom
)


class Group(DataContainer):
    """
    Dictionary-like object to store SPHInX inputs.

    Attributes (sub-groups, parameters, & flags) can be set
    and accessed via dot notation, or as standard dictionary
    key/values.

    `to_{job_type}` converts the Group to the format
    expected by the given DFT code in its input files.
    """

    def to_sphinx(self, content="__self__", indent=0):
        if content == "__self__":
            content = self

        def format_value(v):
            if isinstance(v, bool):
                return f" = {v};".lower()
            elif isinstance(v, Group):
                if len(v) == 0:
                    return " {}"
                else:
                    return " {\n" + self.to_sphinx(v, indent + 1) + indent * "\t" + "}"
            else:
                if isinstance(v, np.ndarray):
                    v = v.tolist()
                return " = {!s};".format(v)

        line = ""
        for k, v in content.items():
            if isinstance(v, Group) and len(v) > 0 and not v.has_keys():
                for vv in v.values():
                    line += indent * "\t" + str(k) + format_value(vv) + "\n"
            else:
                line += indent * "\t" + str(k) + format_value(v) + "\n"

        return line


def get_structure_group(
    positions,
    cell,
    elements,
    movable=None,
    labels=None,
    use_symmetry=True,
    keep_angstrom=False,
):
    """
    create a SPHInX Group object based on structure

    Args:
        positions ((n, 3)-list/numpy.ndarray): xyz-coordinates of the atoms
        cell ((3, 3)-list/numpy.ndarray): Simulation box cdimensions
        elements ((n,)-list/numpy.ndarray): Chemical symbols
        movable (None/(n, 3)-list/nump.ndarray): Whether to fix the
            movement of the atoms in given directions
        labels (None/(n,)-list/numpy.ndarray): Extra labels to distinguish
            atoms for symmetries (mainly for magnetic moments)
        use_symmetry (bool): Whether or not consider internal symmetry
        keep_angstrom (bool): Store distances in Angstroms or Bohr

    Returns:
        (Group): structure group
    """
    positions = np.array(positions)
    cell = np.array(cell)
    if not keep_angstrom:
        cell /= BOHR_TO_ANGSTROM
        positions /= BOHR_TO_ANGSTROM
    structure_group = Group({"cell": np.array(cell)})
    if movable is not None:
        movable = np.array(movable)
    else:
        movable = np.full(shape=positions.shape, fill_value=True)
    if positions.shape != movable.shape:
        raise ValueError("positions.shape != movable.shape")
    if labels is not None:
        labels = np.array(labels)
    else:
        labels = np.full(shape=(len(positions),), fill_value=None)
    if (len(positions),) != labels.shape:
        raise ValueError("len(positions) != labels.shape")
    species = structure_group.create_group("species")
    for elm_species in np.unique(elements):
        species.append(Group({"element": '"' + str(elm_species) + '"'}))
        elm_list = elements == elm_species
        atom_group = species[-1].create_group("atom")
        for elm_pos, elm_magmom, selective in zip(
            positions[elm_list],
            labels[elm_list],
            movable[elm_list],
        ):
            atom_group.append(Group())
            if elm_magmom is not None:
                atom_group[-1]["label"] = '"spin_' + str(elm_magmom) + '"'
            atom_group[-1]["coords"] = np.array(elm_pos)
            if all(selective):
                atom_group[-1]["movable"] = True
            elif any(selective):
                for xx in np.array(["X", "Y", "Z"])[selective]:
                    atom_group[-1]["movable" + xx] = True
    if not use_symmetry:
        structure_group.symmetry = Group(
            {"operator": {"S": "[[1,0,0],[0,1,0],[0,0,1]]"}}
        )
    return structure_group


def copy_potentials(origins, destinations):
    """
    Args:
        origins (list): list of paths to potentials
        destinations (list): list of paths to copy potentials to
    """
    for ori, des in zip(origins, destinations):
        copyfile(ori, des)


def write_spin_constraints(
    file_name="spins.in", cwd=None, magmoms=None, constraints=None
):
    """
    Write a text file containing a list of all spins named spins.in -
    which is used for the external control scripts.

    Args:
        file_name (str): name of the file to be written (optional)
        cwd (str): the current working directory (optinal)
        spins_list (list): list of spins
    """
    for spin in magmoms:
        if isinstance(spin, list) or isinstance(spin, np.ndarray):
            raise ValueError("SPHInX only supports collinear spins at the moment.")
    spins = np.array(magmoms).astype(str)
    spins[~np.asarray(constraints)] = "X"
    spins_str = "\n".join(spins) + "\n"
    if cwd is not None:
        file_name = posixpath.join(cwd, file_name)
    with open(file_name, "w") as f:
        f.write(spins_str)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

__author__ = "Yury Lysogorskiy, Jan Janssen, Marvin Poul"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "Aug 12, 2020"

from pyiron_base import DataContainer, GenericJob
from pyiron_snippets.deprecate import deprecate

from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage


class StructureContainer(GenericJob, HasStructure):
    """
    Container to save a list of structures in HDF5 together with tags.

    Add new structures with :meth:`.append`, they are added to
    :attr:`.structure_lst`.  The HDF5 is written when :meth:`.run` is called.
    """

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.__version__ = "0.2.0"
        self.__hdf_version__ = "0.3.0"
        self._structure_lst = DataContainer(table_name="structures")
        self._container = StructureStorage()
        self.server.run_mode.interactive = True

    @property
    def structure_lst(self):
        """
        :class:`.DataContainer`: list of :class:`~.Atoms`
        """
        if len(self._structure_lst) != len(self._container):
            self._structure_lst = DataContainer(list(self._container.iter_structures()))
        return self._structure_lst

    @staticmethod
    def _to_structure(structure_or_job):
        """
        Return structure from structure or atomic job.

        Args:
            structure_or_job (:class:`~.AtomisticGenericJob`, :class:`~.Atoms`):
                if :class:`~.AtomisticGenericJob` try to get most recent structure,
        copy it and set the job_id in :attr:`~.Atoms.info`

        Returns:
            :class:`~.Atoms`: structure from the job or given structure

        Raises:
            ValueError: if given :class:`~.AtomisticGenericJob` has no structure set
            TypeError: if structure_or_job is of invalid type
        """
        if isinstance(structure_or_job, AtomisticGenericJob):
            if structure_or_job.structure:
                s = structure_or_job.get_structure(-1).copy()
                s.info["jobid"] = structure_or_job.job_id
                return s
            else:
                raise ValueError("The job does not contain any structure to import.")
        elif isinstance(structure_or_job, Atoms):
            return structure_or_job
        else:
            raise TypeError(
                f"structure_or_job must be of type {Atoms} or {AtomisticGenericJob}, not {type(structure_or_job)}"
            )

    def append(self, structure_or_job):
        """
        Add new structure to structure list.

        The added structure will available in :attr:`~.structure_lst`.  If the
        structure is added via a job, retrieve the latest structure and add its
        id to :attr:`pyiron_atomistics.atomistics.generic.Atoms.info`.

        Args:
            structure_or_job (:class:`~.AtomisticGenericJob`/:class:`~.Atoms`):
                if :class:`~.AtomisticGenericJob` add from
                :meth:`~.AtomisticGenericJob.get_structure`,
                otherwise add just the given :class:`~.Atoms`

        Returns:
            dict: item added to :attr:`~.structure_lst`
        """
        struct = self._to_structure(structure_or_job)
        self._container.add_structure(struct)
        return struct

    def add_structure(self, structure: Atoms, identifier: str = None, **kwargs):
        """
        Add a new structure.

        Args:
            structure (:class:`~.Atoms`): structure to add
            identifier (str, optional): optional identifier for the structure
            **kwargs: passed through to the underlying :meth:`.StructureStorage.add_structure`
        """
        self._container.add_structure(structure, identifier=identifier, **kwargs)

    def run_static(self):
        self.status.finished = True

    def run_if_interactive(self):
        self.to_hdf()
        self.status.finished = True

    def write_input(self):
        pass

    def collect_output(self):
        pass

    @property
    @deprecate("use get_structure()")
    def structure(self):
        return self._get_structure(frame=0)

    @structure.setter
    @deprecate("use append()")
    def structure(self, struct):
        self.append(struct)

    def _number_of_structures(self):
        return len(self._container)

    def _get_structure(self, frame=-1, wrap_atoms=True):
        return self._container._get_structure(frame=frame, wrap_atoms=wrap_atoms)

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self._container.to_hdf(hdf=self.project_hdf5, group_name="structures")

    def from_hdf(self, hdf=None, group_name=None):
        # keep hdf structure for version peeking in separate variable, so that
        # the inherited from_hdf() can properly deal with it
        h5 = hdf or self.project_hdf5
        if group_name:
            h5 = h5[group_name]
        if "HDF_VERSION" in h5.list_nodes():
            hdf_version = h5["HDF_VERSION"]
        else:
            # old versions didn't use to set a HDF version
            hdf_version = "0.1.0"
        if hdf_version == "0.1.0":
            super().from_hdf(hdf=hdf, group_name=group_name)
            with self.project_hdf5.open("input") as hdf5_input:
                self.append(Atoms().from_hdf(hdf5_input))
        elif hdf_version == "0.2.0":
            GenericJob.from_hdf(self, hdf=hdf, group_name=group_name)

            hdf = self.project_hdf5["structures"]
            for group in sorted(hdf.list_groups()):
                self.append(Atoms().from_hdf(hdf=hdf, group_name=group))
        else:
            super().from_hdf(hdf=hdf, group_name=group_name)
            self._container.from_hdf(hdf=self.project_hdf5, group_name="structures")

    @property
    def plot(self):
        """
        Accessor for :class:`~.StructurePlots` instance using these structures.
        """
        return self._container.plot



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
import shutil
from typing import List

import pandas as pd
from pyiron_base import GenericParameters, state
from pyiron_snippets.resources import ResourceResolver

from pyiron_atomistics.atomistics.job.potentials import PotentialAbstract
from pyiron_atomistics.atomistics.structure.atoms import Atoms

__author__ = "Joerg Neugebauer, Sudarsan Surendralal, Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class LammpsPotential(GenericParameters):
    """
    This module helps write commands which help in the control of parameters related to the potential used in LAMMPS
    simulations
    """

    def __init__(self, input_file_name=None):
        super(LammpsPotential, self).__init__(
            input_file_name=input_file_name,
            table_name="potential_inp",
            comment_char="#",
        )
        self._potential = None
        self._attributes = {}
        self._df = None

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, new_dataframe):
        self._df = new_dataframe
        # ToDo: In future lammps should also support more than one potential file - that is currently not implemented.
        try:
            self.load_string("".join(list(new_dataframe["Config"])[0]))
        except IndexError:
            raise ValueError(
                "Potential not found! "
                "Validate the potential name by self.potential in self.list_potentials()."
            )

    def remove_structure_block(self):
        self.remove_keys(["units"])
        self.remove_keys(["atom_style"])
        self.remove_keys(["dimension"])

    @property
    def files(self):
        env = os.environ
        if len(self._df["Filename"].values[0]) > 0 and self._df["Filename"].values[
            0
        ] != [""]:
            absolute_file_paths = [
                files for files in list(self._df["Filename"])[0] if os.path.isabs(files)
            ]
            relative_file_paths = [
                files
                for files in list(self._df["Filename"])[0]
                if not os.path.isabs(files)
            ]
            for path in relative_file_paths:
                absolute_file_paths.append(
                    LammpsPotentialFile.find_potential_file(path)
                )
            return absolute_file_paths

    def copy_pot_files(self, working_directory):
        if self.files is not None:
            _ = [shutil.copy(path_pot, working_directory) for path_pot in self.files]

    def get_element_lst(self):
        return list(self._df["Species"])[0]

    def _find_line_by_prefix(self, prefix):
        """
        Find a line that starts with the given prefix.  Differences in white
        space are ignored.  Raises a ValueError if not line matches the prefix.

        Args:
            prefix (str): line prefix to search for

        Returns:
            list: words of the matching line

        Raises:
            ValueError: if not matching line was found
        """

        def isprefix(prefix, lst):
            if len(prefix) > len(lst):
                return False
            return all(n == l for n, l in zip(prefix, lst))

        # compare the line word by word to also match lines that differ only in
        # whitespace
        prefix = prefix.split()
        for parameter, value in zip(self._dataset["Parameter"], self._dataset["Value"]):
            words = (parameter + " " + value).strip().split()
            if isprefix(prefix, words):
                return words

        raise ValueError('No line with prefix "{}" found.'.format(" ".join(prefix)))

    def get_element_id(self, element_symbol):
        """
        Return numeric element id for element. If potential does not contain
        the element raise a :class:NameError.  Only makes sense for potentials
        with pair_style "full".

        Args:
            element_symbol (str): short symbol for element

        Returns:
            int: id matching the given symbol

        Raise:
            NameError: if potential does not contain this element
        """

        try:
            line = "group {} type".format(element_symbol)
            return int(self._find_line_by_prefix(line)[3])

        except ValueError:
            msg = "potential does not contain element {}".format(element_symbol)
            raise NameError(msg) from None

    def get_charge(self, element_symbol):
        """
        Return charge for element. If potential does not specify a charge,
        raise a :class:NameError.  Only makes sense for potentials
        with pair_style "full".

        Args:
            element_symbol (str): short symbol for element

        Returns:
            float: charge speicified for the given element

        Raises:
            NameError: if potential does not specify charge for this element
        """

        try:
            line = "set group {} charge".format(element_symbol)
            return float(self._find_line_by_prefix(line)[4])

        except ValueError:
            msg = "potential does not specify charge for element {}".format(
                element_symbol
            )
            raise NameError(msg) from None

    def to_dict(self):
        super_dict = super(LammpsPotential, self).to_dict()
        if self._df is not None:
            super_dict.update(
                {
                    "potential/" + key: self._df[key].values[0]
                    for key in ["Config", "Filename", "Name", "Model", "Species"]
                }
            )
            if "Citations" in self._df.columns.values:
                super_dict["potential/Citations"] = self._df["Citations"].values[0]
        return super_dict

    def from_dict(self, obj_dict, version: str = None):
        super(LammpsPotential, self).from_dict(obj_dict=obj_dict, version=version)
        if "potential" in obj_dict.keys() and "Config" in obj_dict["potential"].keys():
            entry_dict = {
                key: [obj_dict["potential"][key]]
                for key in ["Config", "Filename", "Name", "Model", "Species"]
            }
            if "Citations" in obj_dict["potential"].keys():
                entry_dict["Citations"] = [obj_dict["potential"]["Citations"]]
            self._df = pd.DataFrame(entry_dict)

    def to_hdf(self, hdf, group_name=None):
        if self._df is not None:
            with hdf.open("potential") as hdf_pot:
                hdf_pot["Config"] = self._df["Config"].values[0]
                hdf_pot["Filename"] = self._df["Filename"].values[0]
                hdf_pot["Name"] = self._df["Name"].values[0]
                hdf_pot["Model"] = self._df["Model"].values[0]
                hdf_pot["Species"] = self._df["Species"].values[0]
                if "Citations" in self._df.columns.values:
                    hdf_pot["Citations"] = self._df["Citations"].values[0]
        super(LammpsPotential, self).to_hdf(hdf, group_name=group_name)

    def from_hdf(self, hdf, group_name=None):
        with hdf.open("potential") as hdf_pot:
            try:
                entry_dict = {
                    "Config": [hdf_pot["Config"]],
                    "Filename": [hdf_pot["Filename"]],
                    "Name": [hdf_pot["Name"]],
                    "Model": [hdf_pot["Model"]],
                    "Species": [hdf_pot["Species"]],
                }
                if "Citations" in hdf_pot.list_nodes():
                    entry_dict["Citations"] = [hdf_pot["Citations"]]
                self._df = pd.DataFrame(entry_dict)
            except ValueError:
                pass
        super(LammpsPotential, self).from_hdf(hdf, group_name=group_name)


class LammpsPotentialFile(PotentialAbstract):
    """
    The Potential class is derived from the PotentialAbstract class, but instead of loading the potentials from a list,
    the potentials are loaded from a file.

    Args:
        potential_df:
        default_df:
        selected_atoms:
    """

    resource_plugin_name = "lammps"

    @classmethod
    def _get_resolver(cls):
        env = os.environ
        return (
            super()
            ._get_resolver()
            .chain(
                ResourceResolver(
                    [env[var] for var in ("CONDA_PREFIX", "CONDA_DIR") if var in env],
                    "share",
                    "iprpy",
                )
            )
        )

    def __init__(self, potential_df=None, default_df=None, selected_atoms=None):
        if potential_df is None:
            potential_df = self._get_potential_df(
                file_name_lst={"potentials_lammps.csv"},
            )
        super(LammpsPotentialFile, self).__init__(
            potential_df=potential_df,
            default_df=default_df,
            selected_atoms=selected_atoms,
        )
        if len(self.list()) == 0:
            state.logger.warning(
                "It looks like your potential database is empty. In order to"
                " install the standard pyiron library, run:\n\n"
                "conda install -c conda-forge pyiron-data\n\n"
                "Depending on the circumstances, you might have to change the"
                " RESOURCE_PATHS of your .pyiron file. It is typically located in"
                " your home directory. More can be found on the installation page"
                " of the pyiron website."
            )

    def default(self):
        if self._default_df is not None:
            atoms_str = "_".join(sorted(self._selected_atoms))
            return self._default_df[
                (self._default_df["Name"] == self._default_df.loc[atoms_str].values[0])
            ]
        return None

    def find_default(self, element):
        """
        Find the potentials

        Args:
            element (set, str): element or set of elements for which you want the possible LAMMPS potentials
            path (bool): choose whether to return the full path to the potential or just the potential name

        Returns:
            list: of possible potentials for the element or the combination of elements

        """
        if isinstance(element, set):
            element = element
        elif isinstance(element, list):
            element = set(element)
        elif isinstance(element, str):
            element = set([element])
        else:
            raise TypeError("Only, str, list and set supported!")
        element_lst = list(element)
        if self._default_df is not None:
            merged_lst = list(set(self._selected_atoms + element_lst))
            atoms_str = "_".join(sorted(merged_lst))
            return self._default_df[
                (self._default_df["Name"] == self._default_df.loc[atoms_str].values[0])
            ]
        return None

    def __getitem__(self, item):
        potential_df = self.find(element=item)
        selected_atoms = self._selected_atoms + [item]
        return LammpsPotentialFile(
            potential_df=potential_df,
            default_df=self._default_df,
            selected_atoms=selected_atoms,
        )


class PotentialAvailable(object):
    def __init__(self, list_of_potentials):
        self._list_of_potentials = {
            "pot_" + v.replace("-", "_").replace(".", "_"): v
            for v in list_of_potentials
        }

    def __getattr__(self, name):
        if name in self._list_of_potentials.keys():
            return self._list_of_potentials[name]
        else:
            raise AttributeError

    def __dir__(self):
        return list(self._list_of_potentials.keys())

    def __repr__(self):
        return str(dir(self))


def view_potentials(structure: Atoms) -> pd.DataFrame:
    """
    List all interatomic potentials for the given atomistic structure including all potential parameters.

    To quickly get only the names of the potentials you can use `list_potentials()` instead.

    Args:
        structure (Atoms): The structure for which to get potentials.

    Returns:
        pandas.Dataframe: Dataframe including all potential parameters.
    """
    list_of_elements = set(structure.get_chemical_symbols())
    return LammpsPotentialFile().find(list_of_elements)


def list_potentials(structure: Atoms) -> List[str]:
    """
    List of interatomic potentials suitable for the given atomic structure.

    See `view_potentials` to get more details.

    Args:
        structure (Atoms): The structure for which to get potentials.

    Returns:
        list: potential names
    """
    return list(view_potentials(structure)["Name"].values)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import numpy as np
from pyiron_base import JobGenerator

from pyiron_atomistics.atomistics.master.parallel import AtomisticParallelMaster

try:
    import pylab as plt
except (ImportError, RuntimeError):
    try:
        import matplotlib.pyplot as plt
    except (ImportError, RuntimeError):
        pass

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class EncutConvergenceJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        return [
            np.round(encut, 7)
            for encut in np.linspace(
                self._master.input["min"],
                self._master.input["max"],
                self._master.input["num_points"],
            )
        ]

    @staticmethod
    def job_name(parameter):
        return "encut_" + str(parameter).replace(".", "_")

    @staticmethod
    def modify_job(job, parameter):
        job.set_encut(encut=parameter)
        return job


# ToDo: not all abstract methods implemented
class ConvEncutParallel(AtomisticParallelMaster):
    def __init__(self, project, job_name="encut_conv"):
        """

        Args:
            project:
            job_name:
        """
        super(ConvEncutParallel, self).__init__(project, job_name)

        self.__version__ = "0.0.1"

        # define default input
        self.input["num_points"] = (11, "number of sample points")
        self.input["min"] = (200, "EnCut Minimum")
        self.input["max"] = (800, "EnCut Maximum")
        self._job_generator = EncutConvergenceJobGenerator(self)

    def collect_output(self):
        eng_lst, encut_lst = [], []
        for job_id in self.child_ids:  # add iter_jobs (should behave like a project)
            ham = self.project_hdf5.inspect(job_id)
            print("job_id: ", job_id, ham.status)
            eng_lst.append(ham["output/generic/energy_tot"][-1])
            encut = ham.job_name.split("_")[1:]
            encut_lst.append(float(encut[0] + "." + encut[1]))
        arg_lst = np.argsort(encut_lst)
        self._output["energy"] = eng_lst[arg_lst]
        self._output["encut"] = encut_lst[arg_lst]

        with self.project_hdf5.open("output") as hdf5_out:
            for key, val in self._output.items():
                hdf5_out[key] = val

    def plot(self, plt_show=True):
        df = self.output_to_pandas()
        encut_lst, eng_lst = df["encut"], df["energy"]
        plt.plot(encut_lst, eng_lst, "x-", markersize=20)
        plt.xlabel("EnCut (eV)")
        plt.ylabel("energy (eV)")
        if plt_show:
            plt.show()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_vasp.vasp.volumetric_data import (
    VaspVolumetricData as _VaspVolumetricData,
)

from pyiron_atomistics.atomistics.structure.atoms import ase_to_pyiron

__author__ = "Sudarsan Surendralal"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class VaspVolumetricData(_VaspVolumetricData):
    def to_hdf(self, hdf, group_name="volumetric_data"):
        """
        Writes the data as a group to a HDF5 file

        Args:
            hdf (pyiron_base.generic.hdfio.ProjectHDFio): The HDF file/path to write the data to
            group_name (str): The name of the group under which the data must be stored as

        """
        volumetric_data_dict_to_hdf(
            data_dict=self.to_dict(),
            hdf=hdf,
            group_name=group_name,
        )

    def from_hdf(self, hdf, group_name="volumetric_data"):
        """
        Recreating the VolumetricData instance by reading data from the HDF5 files

        Args:
            hdf (pyiron_base.generic.hdfio.ProjectHDFio): The HDF file/path to write the data to
            group_name (str): The name of the group under which the data must be stored as

        Returns:
            pyiron.atomistics.volumetric.generic.VolumetricData: The VolumetricData instance

        """
        with hdf.open(group_name) as hdf_vd:
            self._total_data = hdf_vd["total"]
            if "diff" in hdf_vd.list_nodes():
                self._diff_data = hdf_vd["diff"]

    def _read_vol_data_old(self, filename, normalize=True):
        """
        Convenience method to parse a generic volumetric static file in the vasp like format.
        Used by subclasses for parsing the file. This routine is adapted from the pymatgen vasp VolumetricData
        class with very minor modifications. The new parser is faster

        http://pymatgen.org/_modules/pymatgen/io/vasp/outputs.html#VolumetricData.

        Args:
            filename (str): Path of file to parse
            normalize (boolean): Flag to normalize by the volume of the cell

        """
        atoms, total_data_list = super()._read_vol_data_old(
            filename=filename, normalize=normalize
        )
        if atoms is not None:
            atoms = ase_to_pyiron(atoms)
        return atoms, total_data_list

    def _read_vol_data(self, filename, normalize=True):
        """
        Parses the VASP volumetric type files (CHGCAR, LOCPOT, PARCHG etc). Rather than looping over individual values,
        this function utilizes numpy indexing resulting in a parsing efficiency of at least 10%.

        Args:
            filename (str): File to be parsed
            normalize (bool): Normalize the data with respect to the volume (Recommended for CHGCAR files)

        Returns:
            pyiron.atomistics.structure.atoms.Atoms: The structure of the volumetric snapshot
            list: A list of the volumetric data (length >1 for CHGCAR files with spin)

        """
        atoms, total_data_list = super()._read_vol_data(
            filename=filename, normalize=normalize
        )
        if atoms is not None:
            atoms = ase_to_pyiron(atoms)
        return atoms, total_data_list


def volumetric_data_dict_to_hdf(data_dict, hdf, group_name="volumetric_data"):
    with hdf.open(group_name) as hdf_vd:
        for k, v in data_dict.items():
            hdf_vd[k] = v



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

import numpy as np

if TYPE_CHECKING:
    from pyiron_atomistics.atomistics.structure.atoms import Atoms


def remap_indices(
    lammps_indices: Union[np.ndarray, List],
    potential_elements: Union[np.ndarray, List],
    structure: Atoms,
) -> np.ndarray:
    """
    Give the Lammps-dumped indices, re-maps these back onto the structure's indices to preserve the species.

    The issue is that for an N-element potential, Lammps dumps the chemical index from 1 to N based on the order
    that these species are written in the Lammps input file. But the indices for a given structure are based on the
    order in which chemical species were added to that structure, and run from 0 up to the number of species
    currently in that structure. Therefore we need to be a little careful with mapping.

    Args:
        lammps_indices (numpy.ndarray/list): The Lammps-dumped integers.
        potential_elements (numpy.ndarray/list):
        structure (pyiron_atomistics.atomistics.structure.Atoms):

    Returns:
        numpy.ndarray: Those integers mapped onto the structure.
    """
    lammps_symbol_order = np.array(potential_elements)

    # If new Lammps indices are present for which we have no species, extend the species list
    unique_lammps_indices = np.unique(lammps_indices)
    if len(unique_lammps_indices) > len(np.unique(structure.indices)):
        unique_lammps_indices -= (
            1  # Convert from Lammps start counting at 1 to python start counting at 0
        )
        new_lammps_symbols = lammps_symbol_order[unique_lammps_indices]
        structure.set_species(
            [structure.convert_element(el) for el in new_lammps_symbols]
        )

    # Create a map between the lammps indices and structure indices to preserve species
    structure_symbol_order = np.array([el.Abbreviation for el in structure.species])
    map_ = np.array(
        [
            int(np.argwhere(lammps_symbol_order == symbol)[0]) + 1
            for symbol in structure_symbol_order
        ]
    )

    structure_indices = np.array(lammps_indices)
    for i_struct, i_lammps in enumerate(map_):
        np.place(structure_indices, lammps_indices == i_lammps, i_struct)
    # TODO: Vectorize this for-loop for computational efficiency

    return structure_indices
